/*
 * Copyright (c) 2020-2022, NVIDIA CORPORATION.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <cugraph-ops/utils/error.hpp>
#include <cugraph-ops/utils/logger.hpp>

namespace cugraph::ops::gnn {

/**
 * @brief Type of MMA OPeration
 *
 */
enum class MMAOpT {
  /** use high precision tensor core MMA operation
      for fp32, this means using the 3x TF32 "trick"
  */
  kHighPrecision = 0,
  /** use low precision tensor core MMA operation
      for fp32, this means using a single TF32 MMA
      for fp64, this is equivalent to kHighPrecision
  */
  kLowPrecision,
  /** use a simple kSimt MM implementation */
  kSimt
};  // enum class MMAOptT

}  // namespace cugraph::ops::gnn
