/*
 * Copyright (c) 2019-2022, NVIDIA CORPORATION.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

namespace cugraph::ops::gnn::utils {

/** at runtime disable the creation of nvtx ranges */
void disable_nvtx_ranges();

/** at runtime enable the creation of nvtx ranges (default is, disabled) */
void enable_nvtx_ranges();

/**
 * @brief Push a named nvtx range
 * @param name range name
 */
void push_range(const char* name);

/** Pop the latest range */
void pop_range();

}  // namespace cugraph::ops::gnn::utils
