/*
 * Copyright (c) 2020-2022, NVIDIA CORPORATION.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

namespace cugraph::ops::gnn {

/** Types of activations supported */
enum class ActivationOpT {
  /** y = x */
  kLinear = 0,
  /** y = relu(x) */
  kRelu,
  /** y = 1 / (e^-x + 1) */
  kSigmoid,
  /** y = tanh(x) */
  kTanh,
  /** y = e^x - 1 if x < 0 else x */
  kELU,
  /** y = alpha * x */
  kScalar,
  /** y = x * alpha if x < 0 else x */
  kLeakyRelu,
};  // enum ActivationOpT

/**
 * @brief a god(!) struct to store all activation params
 */
struct activation_params {
  /** activation type */
  ActivationOpT type{ActivationOpT::kLinear};
  /** negative slope for the leaky-relu */
  float alpha{0.F};
};  // struct activation_params

}  // namespace cugraph::ops::gnn
