/*
 * Copyright (c) 2020-2022, NVIDIA CORPORATION.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <cugraph-ops/graph/format.hpp>

#include <cstdint>

namespace cugraph::ops::gnn {

/**
 * @brief Computes the max vertex degree of the input graph
 *
 * @param[out] out    will contain the max vertex degree [on device] [len = `1`]
 * @param[in]  graph  input graph whose max vertex degree needs to be computed
 * @param[in]  stream cuda stream where to schedule work
 *
 * @{
 */
void max_vertex_degree(int32_t* out, const fg_csr_s32_t& graph, cudaStream_t stream);
void max_vertex_degree(int64_t* out, const fg_csr_s64_t& graph, cudaStream_t stream);
/** @} */

}  // namespace cugraph::ops::gnn
